require 'delegate'
require 'yaml'

class YAMLSettings < DelegateClass(Hash)

  VERSION = '1.0.6'

  module AttrReader # :nodoc:
    def method_missing(setting)
      has_key?(setting) ? fetch(setting) : fetch(setting.to_s) # raises KeyError
    end
    private :method_missing
  end

  #
  # Settings are merged in two stages. First, all +filenames+ are recursively
  # merged in reverse order. Settings in succeeding +filenames+ overwrite all
  # duplicate settings in preceding +filenames+. Second, settings specified by
  # +key+ are then recursively merged with the _default_ settings.
  #
  # +filenames+:: one or more YAML data files
  # +key+::       selects the top-level stanza from +filenames+ to be merged
  #               with the _default_ stanza
  #
  def initialize(*filenames, key)
    all_settings = filenames.collect {|f| YAML.load_file(f) }
    settings = all_settings.reduce {|i, j| deep_merge(i, j) }
    @settings = deep_merge(settings.fetch('default'),
                           settings.fetch(key.to_s))
    @settings.extend(AttrReader)
    propagate_attr_reader(@settings)
    super(@settings)
  end

  def inspect() @settings.inspect end
  alias :to_s :inspect

  private

  def method_missing(setting, *args, &blk)
    super rescue @settings.send(setting)
  end

  def deep_merge(h1, h2)
    h1.merge!(h2) do |key, v1, v2|
      v1.is_a?(Hash) && v2.is_a?(Hash) ? deep_merge(v1, v2) : v2
    end
  end

  def propagate_attr_reader(settings)
    settings.each_value do |setting_value|
      if setting_value.is_a? Hash
        setting_value.extend(AttrReader)
        propagate_attr_reader(setting_value)
      end
    end
  end

end
